/* ---------------------------------------------------------------------------------
 * Rseau en anneau  jeton (token ring) : une station de l'anneau
 * Jean-Paul Rigault 
 * ---------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "station.h"

/* Variables globales
 * ------------------
 */
FILE *fp_in;	 	/* Pour lecture au terminal */
FILE *fp_out;	 	/* Pour criture au terminal */
int this_station;	/* Numro de cette station */

/* Dclarations en avant
 * ---------------------
*/

static char *message_to_transmit();	/* Demande de message  transmettre */

static void trace_received(const struct packet *ppack);    /* Rception d'un message */
static void trace_acknowledged(const struct packet *ppack);/* Rception accus de rception */
static void trace_passing(const struct packet *ppack);     /* Propagation simple */
static void trace_emitting(const struct packet *ppack);    /* mission d'un message */

static void trace(const char *msg)
{
    fprintf(stderr, "*** %s\n", msg);
}

/* Programme principal
 * -------------------
 *
 * La station est connecte  ses deux voisines par l'intermdiaire de 'stdin'
 * (0) et 'stdout' (1); elle est galement capable de lire et d'crire sur le
 * terminal.
 * 									
 * L'algorithme est le suivant:
 * 									
 * 	1. la station se bloque en lecture de paquet sur son entre standard.
 * 									
 *	2. lorsque la lecture se dbloque et que le jeton est libre, la station
 *	demande sur le terminal si il y a un message  envoyer et, si oui, 
 *	quelle station et quel est son contenu. Dans l'affirmative elle compose
 *	le paquet et l'envoie sur l'anneau; sinon, elle propage le paquet avec
 *	le jeton libre sur l'anneau.
 *	
 * 	3. si le jeton n'est pas libre, c'est qu'il y a des donnes valides qui
 * 	suivent. Trois cas sont alors possibles :
 * 									
 *	    3.1. La station est destinatrice des donnes : elle les traite
 *	    (ici elle se contente de les afficher), marque qu'elle les a reues
 *	    correctement (dans le champ accus de rception du paquet) et
 *	    retransmet le paquet sur le rseau pour qu'il puisse revenir  la
 *	    station mettrice.
 * 									
 * 	    3.2. La station est l'mettrice du paquet : ceci signifie que le
 * 	    message a fait un tour complet.  La station peut alors dterminer
 * 	    par l'analyse  de l'accus de rception si le destinataire a
 * 	    bien recu le message. Dans tous les cas, la station doit librer
 * 	    le jeton, et transmettre un paquet  jeton libre sur l'anneau.
 * 									
 *	    3.3. Aucun des 2 cas prcdents : le paquet ne concerne pas la
 *	    station qui le propage tel quel sur l'anneau.
 * 									
 * C'est la station 0 qui injecte le premier paquet (avec un jeton libre) sur
 * le rseau.
*/ 
int main(int argc, char *argv[])
{
    struct packet current_packet;	/* Le paquet courant */
    int ndest;				/* la station destinatrice */
    char *msg;				/* le message  transmettre */
    int n;
        
    /* Initialisation du numro de la station courante  */
    if (argc != 2)
    {
        fprintf(stderr, "Usage : station.exe numro\n");
        exit(1);
    }
    this_station = atoi(argv[1]);
 
    /* Ouverture du terminal utilisateur pour les traces */
    fp_in = fopen("/dev/tty", "r");
    fp_out = fopen("/dev/tty", "w");
    if (fp_in == NULL || fp_out == NULL) 
    {
	fprintf(stderr, "Erreur station %d: malaise avec le terminal\n",
                this_station);
	exit(1);
    }  
    setbuf(fp_out, NULL);   /* vite la bufferisation en sortie */
       
    /* La station 0 lance la ronde */
    if (this_station == 0)
    {
	current_packet.m_token = TOK_FREE;	/* jeton libre */
	write(1, &current_packet, SIZE_PACKET);
    }
    
    /* Boucle infinie de gestion du rseau et des paquets */
    while(1) 
    {
	/* La station se bloque, en attente de paquet sur stdin */
	n = read(0, &current_packet, SIZE_PACKET);
 
	/* Un paquet est arriv */
        assert(n == SIZE_PACKET);
	if (current_packet.m_token != TOK_FREE)
	{	    
	    /* Le jeton n'est pas libre : il y a un message joint */
	    if (current_packet.m_dest != this_station
                && current_packet.m_orig != this_station)
	    {
		/* Ce paquet ne nous concerne pas ; on le passe au voisin */

		trace_passing(&current_packet);
	    }
	    else
	    {
                /* Ce paquet nous concerne
                 *
                 * Attention: il faut traiter correctement le cas o c'est un
                 * message  nous-mme qui a fait un tout complet !
                 */

		if (current_packet.m_dest == this_station) 
		{
		    /* ce paquet nous est destin */
		    trace_received(&current_packet);
		    current_packet.m_token = TOK_ACK; /* accus de rception */
		}
		if (current_packet.m_orig == this_station) {
		    /* Nous avons mis ce message et il a fait un tour complet */
		    trace_acknowledged(&current_packet);
		    current_packet.m_token = TOK_FREE;	/* libre le jeton */
		}
	    }
	}
	else if ((msg = message_to_transmit(&ndest)) != NULL)
	{
	    /* le jeton est libre et nous voulons transmettre un message*/

	    strcpy(current_packet.m_message, msg);
	    current_packet.m_token = TOK_BUSY;		/* prend le jeton */
	    current_packet.m_dest = ndest;
	    current_packet.m_orig = this_station;
	    trace_emitting(&current_packet);
	}
	
	/* Dans tous les cas on entretient le rseau en propageant le paquet */

	write(1, &current_packet, SIZE_PACKET);
    }
    /*NOTREACHED*/
}

/* Demande du message a transmettre au terminal		
 * ---------------------------------------------
 * La destination est renvoye dans l'argument pdest, qui est en fait un entier
 * pass par rfrence. La valeur de retour de message_to_transmit(...) est un
 * pointeur sur les nouvelles donnes du message (qui sont ranges dans un
 * tableau local  la fonction).
 * 									
 */
static char *message_to_transmit(int *pdest)
{
    char rep[10];
    static char msg[MSG_LEN];
    int ndest;
    
    trace("message_to_transmit");
    fprintf(fp_out, "ici station %d : avez-vous un message  transmettre ? [y/n] ",
            this_station);
    fscanf(fp_in, "%s", rep);

    switch (rep[0]) 
    {
    case 'y':
    case 'Y':
    case 'o':
    case 'O':
	fprintf(fp_out, "station %d : lequel? ", this_station);
	fscanf(fp_in, "%s", msg);
	fprintf(fp_out, "station %d : a quelle station ? ", this_station);
	fscanf(fp_in, "%d", &ndest);
	*pdest = ndest;
	return msg;
    default:
        return NULL;
    }
    /*NOTREACHED*/
}

/* Fonctions de trace				
 * ------------------
 */

/* La station vient de recevoir un message valide qui lui tait adress */
static void trace_received(const struct packet *ppack)
{
    fprintf(fp_out, "station %d : reu de station %d : %s\n",
	    this_station, ppack->m_orig, ppack->m_message);
}

/* Le dernier message transmis par la station a fait un tour complet : on analyse
 * l'accus de reception pour savoir si le destinataire l'a correctement reu
 */
static void trace_acknowledged(const struct packet *ppack)
{
    fprintf(fp_out, "station %d : envoi  station %d reu %s : %s\n",
	    this_station, ppack->m_dest,
            (ppack->m_token == TOK_ACK) ? "correctement" : "incorrectement",
	    ppack->m_message);
}

/* Un message valide transite juste  travers la station */
static void trace_passing(const struct packet *ppack)
{
    fprintf(fp_out, "station %d : passage de paquet de station %d  %d: %s\n",
	    this_station, ppack->m_orig, ppack->m_dest, ppack->m_message);
}

/* La station transmet un nouveau message */ 
static void trace_emitting(const struct packet *ppack)
{
    fprintf(fp_out, "station %d : mission d'un message pour station %d: %s\n",
	    this_station, ppack->m_dest, ppack->m_message);
}

